#! /usr/bin/python3
import io
import logging

from .event import Event, ExpectMsg, ResolvableBool
from .errors import SpecFileError, EventError
from .namespace import namespace
from pyln.proto.message import Message
from typing import Union, List, Optional, TYPE_CHECKING, cast

if TYPE_CHECKING:
    # Otherwise a circular dependency
    from .runner import Runner, Conn

# These can all be fed to a Sequence() initializer.
SequenceUnion = Union["Sequence", List[Event], Event]


class Sequence(Event):
    """A sequence of ordered events"""

    def __init__(
        self,
        events: Union["Sequence", List[Event], Event],
        enable: ResolvableBool = True,
    ):
        """Events can be a Sequence, a single Event, or a list of Events.  If
        enable is False, this turns into a noop (e.g. if runner doesn't support
        it)."""
        super().__init__()
        self.enable = enable
        if type(events) is Sequence:
            # mypy gets upset because Sequence isn't defined yet.
            self.events = events.events  # type: ignore
            self.enable = events.enable  # type: ignore
            self.name = events.name  # type: ignore
        elif isinstance(events, Event):
            self.events = [events]
        else:
            self.events = events

    def enabled(self, runner: "Runner") -> bool:
        return self.resolve_arg("enable", runner, self.enable)

    def action(self, runner: "Runner", skip_first: bool = False) -> bool:
        super().action(runner)
        all_done = True
        for e in self.events:
            logging.debug(f"receiving event {e}")
            if not e.enabled(runner):
                continue
            if skip_first:
                skip_first = False
            else:
                all_done &= e.action(runner)
        return all_done

    @staticmethod
    def ignored_by_all(
        msg: Message, sequences: List["Sequence"]
    ) -> Optional[List[Message]]:
        # If they all say the same thing, that's the answer.
        rets = [cast(ExpectMsg, s.events[0]).ignore(msg) for s in sequences]
        if all([ignored == rets[0] for ignored in rets[1:]]):
            return rets[0]
        return None

    @staticmethod
    def match_which_sequence(
        runner: "Runner", msg: Message, sequences: List["Sequence"]
    ) -> Optional["Sequence"]:
        """Return which sequence expects this msg, or None"""

        for s in sequences:
            failreason = cast(ExpectMsg, s.events[0]).message_match(runner, msg)
            if failreason is None:
                return s

        return None


class OneOf(Event):
    """Event representing multiple possible sequences, one of which should happen"""

    def __init__(self, *args: SequenceUnion):
        super().__init__()
        self.sequences = []
        for s in args:
            seq = Sequence(s)
            if len(seq.events) == 0:
                raise ValueError("{} is an empty sequence".format(s))
            self.sequences.append(seq)

    def enabled_sequences(self, runner: "Runner") -> List[Sequence]:
        """Returns all enabled sequences"""
        return [s for s in self.sequences if s.enabled(runner)]

    def action(self, runner: "Runner") -> bool:
        super().action(runner)

        # Check they all use the same conn!
        conn: Optional[Conn] = None
        for s in self.sequences:
            c = cast(ExpectMsg, s.events[0]).find_conn(runner)
            if conn is None:
                conn = c
            elif c != conn:
                raise SpecFileError(self, "sequences do not all use the same conn?")
        assert conn

        while True:
            event = self.sequences[0].events[0]
            binmsg = runner.get_output_message(conn, event)
            if binmsg is None:
                raise EventError(self, f"Did not receive a message {event} from runner")

            try:
                msg = Message.read(namespace(), io.BytesIO(binmsg))
            except ValueError as ve:
                raise EventError(self, "Invalid msg {}: {}".format(binmsg.hex(), ve))

            ignored = Sequence.ignored_by_all(msg, self.enabled_sequences(runner))
            # If they gave us responses, send those now.
            if ignored is not None:
                for msg in ignored:
                    binm = io.BytesIO()
                    msg.write(binm)
                    runner.recv(self, conn, binm.getvalue())
                continue

            seq = Sequence.match_which_sequence(
                runner, msg, self.enabled_sequences(runner)
            )
            if seq is not None:
                # We found the sequence, run it
                return seq.action(runner, skip_first=True)

            raise EventError(
                self,
                "None of the sequences {} matched {}".format(
                    self.enabled_sequences(runner), msg.to_str()
                ),
            )


class AnyOrder(Event):
    """Event representing multiple sequences, all of which should happen, but not defined which order they would happen"""

    def __init__(self, *args: SequenceUnion):
        super().__init__()
        self.sequences = []
        for s in args:
            seq = Sequence(s)
            if len(seq.events) == 0:
                raise ValueError("{} is an empty sequence".format(s))
            self.sequences.append(seq)

    def enabled_sequences(self, runner: "Runner") -> List[Sequence]:
        """Returns all enabled sequences"""
        return [s for s in self.sequences if s.enabled(runner)]

    def action(self, runner: "Runner") -> bool:
        super().action(runner)

        # Check they all use the same conn!
        conn = None
        for s in self.sequences:
            c = cast(ExpectMsg, s.events[0]).find_conn(runner)
            if conn is None:
                conn = c
            elif c != conn:
                raise SpecFileError(self, "sequences do not all use the same conn?")
        assert conn

        all_done = True
        sequences = self.enabled_sequences(runner)
        while sequences != []:
            # Get message
            binmsg = runner.get_output_message(conn, sequences[0].events[0])
            if binmsg is None:
                raise EventError(
                    self,
                    "Did not receive a message from runner, still expecting {}".format(
                        [s.events[0] for s in sequences]
                    ),
                )

            try:
                msg = Message.read(namespace(), io.BytesIO(binmsg))
            except ValueError as ve:
                raise EventError(self, "Invalid msg {}: {}".format(binmsg.hex(), ve))

            ignored = Sequence.ignored_by_all(msg, self.enabled_sequences(runner))
            # If they gave us responses, send those now.
            if ignored is not None:
                for msg in ignored:
                    binm = io.BytesIO()
                    msg.write(binm)
                    runner.recv(self, conn, binm.getvalue())
                continue

            seq = Sequence.match_which_sequence(runner, msg, sequences)
            if seq is not None:
                sequences.remove(seq)
                all_done &= seq.action(runner, skip_first=True)
                continue

            raise EventError(
                self,
                "Message did not match any sequences {}: {}".format(
                    [s.events[0] for s in sequences], msg.to_str()
                ),
            )
        return all_done


class TryAll(Event):
    """Event representing multiple sequences, each of which should be tested"""

    def __init__(self, *args: SequenceUnion):
        super().__init__()
        self.sequences = [Sequence(s) for s in args]
        self.done = [False] * len(self.sequences)

    def action(self, runner: "Runner") -> bool:
        super().action(runner)

        # Take first undone one, or if that fails, first enabled one.
        first_enabled = None
        first_undone = None
        all_done = True
        for i, s in enumerate(self.sequences):
            if not s.enabled(runner):
                continue
            if not first_enabled:
                first_enabled = s
            if self.done[i]:
                continue
            if not first_undone:
                first_undone = s
                self.done[i] = True
            else:
                all_done = False

        # Note: they might *all* be disabled!
        if first_undone:
            first_undone.action(runner)
        elif first_enabled:
            first_enabled.action(runner)

        return all_done


def test_empty_sequence() -> None:
    class nullrunner(object):
        class dummyconfig(object):
            def getoption(self, name: str) -> bool:
                return False

        def __init__(self) -> None:
            self.config = self.dummyconfig()

    # This sequence should be tried twice.
    seq = Sequence(TryAll([], []))
    assert seq.action(nullrunner()) is False  # type: ignore
    assert seq.action(nullrunner()) is True  # type: ignore



import functools
import time
import coincurve

from typing import Callable, Optional, Any
from pyln.proto.message import Message

from lnprototest import Runner, Event, Side, SpecFileError, Funding


def commitsig_to_send() -> Callable[[Runner, Event, str], str]:
    """Get the appropriate signature for the local side to send to the remote"""

    def _commitsig_to_send(runner: Runner, event: Event, field: str) -> str:
        tx = runner.get_stash(event, "Commit").remote_unsigned_tx()
        return runner.get_stash(event, "Commit").local_sig(tx)

    return _commitsig_to_send


def commitsig_to_recv() -> Callable[[Runner, Event, str], str]:
    """Get the appropriate signature for the remote side to send to the local"""

    def _commitsig_to_recv(runner: Runner, event: Event, field: str) -> str:
        tx = runner.get_stash(event, "Commit").local_unsigned_tx()
        return runner.get_stash(event, "Commit").remote_sig(tx)

    return _commitsig_to_recv


def _htlc_sigs(signer: Side, runner: Runner, event: Event, field: str) -> str:
    sigs = runner.get_stash(event, "Commit").htlc_sigs(signer, not signer)
    return "[" + ",".join([sig.to_str() for sig in sigs]) + "]"


def htlc_sigs_to_send() -> Callable[[Runner, Event, str], str]:
    """Get the HTLC signatures for local side to send to the remote"""
    return functools.partial(_htlc_sigs, Side.local)


def htlc_sigs_to_recv() -> Callable[[Runner, Event, str], str]:
    """Get the HTLC signatures for remote side to send to the local"""
    return functools.partial(_htlc_sigs, Side.remote)


def channel_id() -> Callable[[Runner, Event, str], str]:
    """Get the channel_id for the current Commit"""

    def _channel_id(runner: Runner, event: Event, field: str) -> str:
        return runner.get_stash(event, "Commit").funding.channel_id()

    return _channel_id


def channel_id_v2() -> Callable[[Runner, Event, str], str]:
    """Get the channel_id for the current Commit for a v2 channel open"""

    def _channel_id(runner: Runner, event: Event, field: str) -> str:
        return runner.get_stash(event, "Commit").channel_id_v2()

    return _channel_id


def channel_announcement(
    short_channel_id: str, features: bytes
) -> Callable[[Runner, Event, str], str]:
    """Get the channel_announcement for the current Commit"""

    def _channel_announcement(
        short_channel_id: str, features: bytes, runner: Runner, event: Event, field: str
    ) -> Message:
        return runner.get_stash(event, "Commit").channel_announcement(
            short_channel_id, features
        )

    return functools.partial(_channel_announcement, short_channel_id, features)


def channel_update(
    short_channel_id: str,
    side: Side,
    disable: bool,
    cltv_expiry_delta: int,
    htlc_minimum_msat: int,
    fee_base_msat: int,
    fee_proportional_millionths: int,
    htlc_maximum_msat: Optional[int],
    timestamp: Optional[int] = None,
) -> Callable[[Runner, Event, str], str]:
    """Get a channel_update for the current Commit"""

    def _channel_update(
        short_channel_id: str,
        side: Side,
        disable: bool,
        cltv_expiry_delta: int,
        htlc_minimum_msat: int,
        fee_base_msat: int,
        fee_proportional_millionths: int,
        timestamp: Optional[int],
        htlc_maximum_msat: Optional[int],
        runner: Runner,
        event: Event,
        field: str,
    ) -> Message:
        """Get the channel_update"""
        if timestamp is None:
            timestamp = int(time.time())
            return runner.get_stash(event, "Commit").channel_update(
                short_channel_id,
                side,
                disable,
                cltv_expiry_delta,
                htlc_maximum_msat,
                fee_base_msat,
                fee_proportional_millionths,
                timestamp,
                htlc_maximum_msat,
            )

    return functools.partial(
        _channel_update,
        short_channel_id,
        side,
        disable,
        cltv_expiry_delta,
        htlc_minimum_msat,
        fee_base_msat,
        fee_proportional_millionths,
        htlc_maximum_msat,
        timestamp,
    )


def get_member(
    event: Event, runner: "Runner", stashname: str, var: str, last: bool = True
) -> str:
    """Get member field from stash for ExpectMsg or Msg.

    If var contains a '.' then we look for that message to extract the field.  If last is True, we get the last message, otherwise the first.
    """
    stash = runner.get_stash(event, stashname)
    if "." in var:
        prevname, _, var = var.partition(".")
    else:
        prevname = ""
    if last:
        seq = reversed(stash)
    else:
        seq = stash

    for name, d in seq:
        if prevname == "" or name == prevname:
            if var not in d:
                raise SpecFileError(
                    event,
                    "{}: {} did not receive a {}".format(stashname, prevname, var),
                )
            return d[var]
    raise SpecFileError(event, "{}: have no prior {}".format(stashname, prevname))


def _get_member(
    stashname: str,
    fieldname: Optional[str],
    casttype: Any,
    # This is the signature which Msg() expects for callable values:
    runner: "Runner",
    event: Event,
    field: str,
) -> Any:
    # If they don't specify fieldname, it's same as this field.
    if fieldname is None:
        fieldname = field
    strval = get_member(event, runner, stashname, fieldname)
    try:
        return casttype(strval)
    except ValueError:
        raise SpecFileError(
            event,
            "{}.{} is {}, not a valid {}".format(
                stashname, fieldname, strval, casttype
            ),
        )


def rcvd(
    fieldname: Optional[str] = None, casttype: Any = str
) -> Callable[[Runner, Event, Any], Any]:
    """Use previous ExpectMsg field (as string)

    fieldname can be [msg].[field] or just [field] for last ExpectMsg

    """
    return functools.partial(_get_member, "ExpectMsg", fieldname, casttype)


def sent(
    fieldname: Optional[str] = None, casttype: Any = str
) -> Callable[[Runner, Event, Any], Any]:
    """Use previous Msg field (as string)

    fieldname can be [msg].[field] or just [field] for last Msg

    """
    return functools.partial(_get_member, "Msg", fieldname, casttype)


def funding_amount() -> Callable[[Runner, Event, str], int]:
    """Get the stashed funding amount"""

    def _funding_amount(runner: Runner, event: Event, field: str) -> int:
        return runner.get_stash(event, "Funding").amount

    return _funding_amount


def funding_pubkey(side: Side) -> Callable[[Runner, Event, str], str]:
    """Get the stashed funding pubkey for side"""

    def _funding_pubkey(side: Side, runner: Runner, event: Event, field: str) -> str:
        return coincurve.PublicKey.from_secret(
            runner.get_stash(event, "Funding").funding_privkeys[side].secret
        )

    return functools.partial(_funding_pubkey, side)


def funding_tx() -> Callable[[Runner, Event, str], str]:
    """Get the funding transaction (as stashed by CreateFunding)"""

    def _funding_tx(runner: Runner, event: Event, field: str) -> str:
        return runner.get_stash(event, "FundingTx")

    return _funding_tx


def funding_txid() -> Callable[[Runner, Event, str], str]:
    """Get the stashed funding transaction id"""

    def _funding_txid(runner: Runner, event: Event, field: str) -> str:
        return runner.get_stash(event, "Funding").txid

    return _funding_txid


def funding() -> Callable[[Runner, Event, str], Funding]:
    """Get the stashed Funding (as stashed by CreateFunding or AcceptFunding)"""

    def _funding(runner: Runner, event: Event, field: str) -> Funding:
        return runner.get_stash(event, "Funding")

    return _funding


def witnesses() -> Callable[[Runner, Event, str], str]:
    """Get the witnesses for the stashed funding tx"""

    def _witnesses(runner: Runner, event: Event, field: str) -> str:
        funding = runner.get_stash(event, "Funding")
        return funding.our_witnesses()

    return _witnesses


def locking_script() -> Callable[[Runner, Event, str], str]:
    def _locking_script(runner: Runner, event: Event, field: str) -> str:
        return runner.get_stash(event, "Funding").locking_script().hex()

    return _locking_script


def funding_close_tx() -> Callable[[Runner, Event, str], str]:
    def _funding_close_tx(runner: Runner, event: Event, field: str) -> str:
        return runner.get_stash(event, "Funding").close_tx()

    return _funding_close_tx


def stash_field_from_event(
    stash_key: str, field_name: Optional[str] = None, dummy_val: Optional[Any] = None
) -> Callable[[Runner, Event, str], str]:
    """Generic stash function to get the information back from a previous event"""

    def _stash_field_from_event(runner: Runner, event: Event, field: str) -> str:
        if runner._is_dummy():
            return dummy_val
        field = field if field_name is None else field_name
        return runner.get_stash(event, stash_key).fields[field]

    return _stash_field_from_event



"""Runner for the core-lightning implementation.

This is adapted from the older testcases, so it is overly prescriptive
of how the node is configured.  A more modern implementation would simply
reach into the node and derive the private keys it's using, rather than
use hacky --dev options to override it.

We could also factor out the bitcoind implementation, which is really
independent.

Important environment variables include TIMEOUT which sets how long we
wait for responses (default, 30 seconds), and LIGHTNING_SRC which indicates
where the binaries are (default ../lightning).

"""

from .clightning import Runner

__all__ = ["Runner"]



#! /usr/bin/python3
# FIXME: clean this up for use as pyln.proto.tx
import coincurve
import hashlib


class KeySet(object):
    def __init__(
        self,
        revocation_base_secret: str,
        payment_base_secret: str,
        htlc_base_secret: str,
        delayed_payment_base_secret: str,
        shachain_seed: str,
    ):

        from .utils import privkey_expand, check_hex

        self.revocation_base_secret = privkey_expand(revocation_base_secret)
        self.payment_base_secret = privkey_expand(payment_base_secret)
        self.htlc_base_secret = privkey_expand(htlc_base_secret)
        self.delayed_payment_base_secret = privkey_expand(delayed_payment_base_secret)
        self.shachain_seed = bytes.fromhex(check_hex(shachain_seed, 64))

    def raw_payment_basepoint(self) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(self.payment_base_secret.secret)

    def payment_basepoint(self) -> str:
        return self.raw_payment_basepoint().format().hex()

    def raw_revocation_basepoint(self) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(self.revocation_base_secret.secret)

    def revocation_basepoint(self) -> str:
        return self.raw_revocation_basepoint().format().hex()

    def raw_delayed_payment_basepoint(self) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(self.delayed_payment_base_secret.secret)

    def delayed_payment_basepoint(self) -> str:
        return self.raw_delayed_payment_basepoint().format().hex()

    def raw_htlc_basepoint(self) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(self.htlc_base_secret.secret)

    def htlc_basepoint(self) -> str:
        return self.raw_htlc_basepoint().format().hex()

    def raw_per_commit_secret(self, n: int) -> coincurve.PrivateKey:
        # BOLT #3:
        # The first secret used:
        #  - MUST be index 281474976710655,
        #    - and from there, the index is decremented.
        if n > 281474976710655:
            raise ValueError("48 bits is all you get!")
        index = 281474976710655 - n

        # BOLT #3:
        # generate_from_seed(seed, I):
        #     P = seed
        #     for B in 47 down to 0:
        #         if B set in I:
        #             flip(B) in P
        #             P = SHA256(P)
        #     return P
        # ```

        # FIXME: This is the updated wording from PR #779
        # Where "flip(B)" alternates the (B mod 8)'th bit of the (B div 8)'th
        # byte of the value.  So, "flip(0) in e3b0..." is "e2b0...", and
        # "flip(10) in "e3b0..." is "e3b4".
        P = bytearray(self.shachain_seed)
        for B in range(47, -1, -1):
            if ((1 << B) & index) != 0:
                P[B // 8] ^= 1 << (B % 8)
                P = bytearray(hashlib.sha256(P).digest())

        return coincurve.PrivateKey(P)

    def per_commit_secret(self, n: int) -> str:
        return self.raw_per_commit_secret(n).secret.hex()

    def raw_per_commit_point(self, n: int) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(self.raw_per_commit_secret(n).secret)

    def per_commit_point(self, n: int) -> str:
        return self.raw_per_commit_point(n).format().hex()


def test_shachain() -> None:
    # BOLT #3:
    # ## Generation Tests
    # name: generate_from_seed 0 final node
    # seed: 0x0000000000000000000000000000000000000000000000000000000000000000
    # I: 281474976710655
    # output: 0x02a40c85b6f28da08dfdbe0926c53fab2de6d28c10301f8f7c4073d5e42e3148
    keyset = KeySet(
        "01",
        "01",
        "01",
        "01",
        "0000000000000000000000000000000000000000000000000000000000000000",
    )
    assert (
        keyset.per_commit_secret(0xFFFFFFFFFFFF - 281474976710655)
        == "02a40c85b6f28da08dfdbe0926c53fab2de6d28c10301f8f7c4073d5e42e3148"
    )

    # BOLT #3:
    # name: generate_from_seed FF final node
    # seed: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    # I: 281474976710655
    # output: 0x7cc854b54e3e0dcdb010d7a3fee464a9687be6e8db3be6854c475621e007a5dc
    keyset = KeySet(
        "01",
        "01",
        "01",
        "01",
        "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
    )
    assert (
        keyset.per_commit_secret(0xFFFFFFFFFFFF - 281474976710655)
        == "7cc854b54e3e0dcdb010d7a3fee464a9687be6e8db3be6854c475621e007a5dc"
    )

    # BOLT #3:
    # name: generate_from_seed FF alternate bits 1
    # seed: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    # I: 0xaaaaaaaaaaa
    # output: 0x56f4008fb007ca9acf0e15b054d5c9fd12ee06cea347914ddbaed70d1c13a528
    keyset = KeySet(
        "01",
        "01",
        "01",
        "01",
        "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
    )
    assert (
        keyset.per_commit_secret(0xFFFFFFFFFFFF - 0xAAAAAAAAAAA)
        == "56f4008fb007ca9acf0e15b054d5c9fd12ee06cea347914ddbaed70d1c13a528"
    )

    # BOLT #3:
    # name: generate_from_seed FF alternate bits 2
    # seed: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    # I: 0x555555555555
    # output: 0x9015daaeb06dba4ccc05b91b2f73bd54405f2be9f217fbacd3c5ac2e62327d31
    keyset = KeySet(
        "01",
        "01",
        "01",
        "01",
        "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
    )
    assert (
        keyset.per_commit_secret(0xFFFFFFFFFFFF - 0x555555555555)
        == "9015daaeb06dba4ccc05b91b2f73bd54405f2be9f217fbacd3c5ac2e62327d31"
    )

    # BOLT #3:
    # name: generate_from_seed 01 last nontrivial node
    # seed: 0x0101010101010101010101010101010101010101010101010101010101010101
    # I: 1
    # output: 0x915c75942a26bb3a433a8ce2cb0427c29ec6c1775cfc78328b57f6ba7bfeaa9c
    keyset = KeySet(
        "01",
        "01",
        "01",
        "01",
        "0101010101010101010101010101010101010101010101010101010101010101",
    )
    assert (
        keyset.per_commit_secret(0xFFFFFFFFFFFF - 1)
        == "915c75942a26bb3a433a8ce2cb0427c29ec6c1775cfc78328b57f6ba7bfeaa9c"
    )



#! /usr/bin/python3
import logging
import traceback
import collections
import os.path
import io
import struct
import time
import json

from typing import Optional, Dict, Union, Callable, Any, List, TYPE_CHECKING, overload

from pyln.proto.message import Message

from .errors import SpecFileError, EventError
from .namespace import namespace
from .signature import Sig
from .bitfield import has_bit
from .utils import check_hex

from bitcoin.core import CTransaction

if TYPE_CHECKING:
    # Otherwise a circular dependency
    from .runner import Runner, Conn


# Type for arguments: either strings, or functions to call at runtime
ResolvableStr = Union[str, Callable[["Runner", "Event", str], str]]
ResolvableInt = Union[int, Callable[["Runner", "Event", str], int]]
ResolvableBool = Union[int, Callable[["Runner", "Event", str], bool]]
Resolvable = Union[Any, Callable[["Runner", "Event", str], Any]]


class Event(object):
    """Abstract base class for events."""

    def __init__(self) -> None:
        # From help(traceback.extract_stack):
        #   Each item in the list is a quadruple (filename,
        #   line number, function name, text), and the entries are in order
        #   from oldest to newest stack frame.
        self.name = "unknown"
        for s in reversed(traceback.extract_stack()):
            # Ignore constructor calls, like this one.
            if s[2] != "__init__":
                self.name = "{}:{}:{}".format(
                    type(self).__name__, os.path.basename(s[0]), s[1]
                )
                break

    def enabled(self, runner: "Runner") -> bool:
        """Returns whether it should be enabled for this run.  Usually True"""
        return True

    def action(self, runner: "Runner") -> bool:
        """action() returns the False if it needs to be called again"""
        if runner.config.getoption("verbose"):
            logging.info("# running {}:".format(self.to_json()))
        return True

    def resolve_arg(self, fieldname: str, runner: "Runner", arg: Resolvable) -> Any:
        """If this is a string, return it, otherwise call it to get result"""
        if callable(arg):
            return arg(runner, self, fieldname)
        else:
            return arg

    def resolve_args(
        self, runner: "Runner", kwargs: Dict[str, Resolvable]
    ) -> Dict[str, Any]:
        """Take a dict of args, replace callables with their return values"""
        ret: Dict[str, str] = {}
        for field, str_or_func in kwargs.items():
            ret[field] = self.resolve_arg(field, runner, str_or_func)
        return ret

    def to_json(self) -> Dict[Any, Any]:
        toks = self.name.split(":")
        event = self.name
        file_name = ""
        pos = ""
        if len(toks) > 0:
            event = toks[0]
            file_name = toks[1]
            pos = toks[2]
        return {
            "event": event,
            "file": file_name,
            "pos": pos,
        }

    def __str__(self) -> str:
        return json.dumps(self.to_json())


class PerConnEvent(Event):
    """An event which takes a connprivkey arg"""

    def __init__(self, connprivkey: Optional[str]):
        super().__init__()
        self.connprivkey = connprivkey

    def find_conn(self, runner: "Runner") -> "Conn":
        """Helper for events which have a connection"""
        conn = runner.find_conn(self.connprivkey)
        if conn is None:
            if self.connprivkey is None:
                # None means "same as last used/created"
                raise SpecFileError(self, "No current connection")
            else:
                raise SpecFileError(
                    self, "Unknown connection {}".format(self.connprivkey)
                )
        return conn


class Connect(Event):
    """Connect to the runner, as if a node with private key connprivkey"""

    def __init__(self, connprivkey: str):
        self.connprivkey = connprivkey
        super().__init__()

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        if runner.find_conn(self.connprivkey):
            raise SpecFileError(
                self, "Already have connection to {}".format(self.connprivkey)
            )
        # This is a hack: if we've already got a connection, wait 1 second
        # for gossip to be processed before connecting another one!
        if len(runner.conns) != 0:
            time.sleep(1)
        runner.connect(self, self.connprivkey)
        return True


class MustNotMsg(PerConnEvent):
    """Indicate that this connection must never send any of these message types."""

    def __init__(self, must_not: str, connprivkey: Optional[str] = None):
        super().__init__(connprivkey)
        self.must_not = must_not

    def matches(self, binmsg: bytes) -> bool:
        msgnum = struct.unpack(">H", binmsg[0:2])[0]
        msgtype = namespace().get_msgtype_by_number(msgnum)
        if msgtype:
            name = msgtype.name
        else:
            name = str(msgnum)
        logging.debug(f"msg {name} != from what we are looking for {self.must_not}?")
        return name == self.must_not

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        self.find_conn(runner).must_not_events.append(self)
        return True


class Disconnect(PerConnEvent):
    """Disconnect the runner from the node whose private key is connprivkey: default is last connection specified"""

    def __init__(self, connprivkey: Optional[str] = None):
        super().__init__(connprivkey)

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        runner.disconnect(self, self.find_conn(runner))
        return True


class Msg(PerConnEvent):
    """Feed a message to the runner (via optional given connection)"""

    def __init__(
        self,
        msgtypename: str,
        connprivkey: Optional[str] = None,
        **kwargs: Union[ResolvableStr, ResolvableInt],
    ):
        super().__init__(connprivkey)
        self.msgtype = namespace().get_msgtype(msgtypename)

        if not self.msgtype:
            raise SpecFileError(self, "Unknown msgtype {}".format(msgtypename))
        self.kwargs = kwargs

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        # Now we have runner, we can fill in all the message fields
        message = Message(self.msgtype, **self.resolve_args(runner, self.kwargs))
        missing = message.missing_fields()
        if missing:
            raise SpecFileError(self, "Missing fields {}".format(missing))
        binmsg = io.BytesIO()
        message.write(binmsg)
        runner.recv(self, self.find_conn(runner), binmsg.getvalue())
        msg_to_stash(runner, self, message)
        return True


class Wait(PerConnEvent):
    """Put a delay in a test, to allow time for things to happen
    on the node's end"""

    def __init__(self, delay_s: int):
        super().__init__(None)
        self.delay_s = delay_s

    def action(self, runner: "Runner") -> bool:
        time.sleep(self.delay_s)
        return True


class RawMsg(PerConnEvent):
    """Feed a raw binary, or raw Message to the runner (via optional given connection)"""

    def __init__(
        self,
        message: Union[Resolvable, bytes, Message],
        connprivkey: Optional[str] = None,
    ):
        super().__init__(connprivkey)
        self.message = message

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        msg = self.resolve_arg("binmsg", runner, self.message)
        if isinstance(msg, Message):
            buf = io.BytesIO()
            msg.write(buf)
            binmsg = buf.getvalue()
        else:
            binmsg = msg

        runner.recv(self, self.find_conn(runner), binmsg)
        return True


class ExpectMsg(PerConnEvent):
    """Wait for a message from the runner.

    Args is the (usually incomplete) message which it should match.
    if_match is the function to call if it matches: should raise an
    exception if it's not satisfied.  ignore function to ignore unexpected
    messages: it returns a list of messages to reply with, or None if the
    message should not be ignored: by default, it is ignore_gossip_queries.

    """

    def _default_if_match(self, msg: Message, runner: "Runner") -> None:
        pass

    @staticmethod
    def ignore_pings(msg: Message) -> Optional[List[Message]]:
        """Function to ignore pings (and respond with pongs appropriately)"""
        if msg.messagetype.name != "ping":
            return None

        # BOLT #1:
        # A node receiving a `ping` message:
        # ...
        #  - if `num_pong_bytes` is less than 65532:
        #    - MUST respond by sending a `pong` message, with `byteslen` equal
        #     to `num_pong_bytes`.
        #  - otherwise (`num_pong_bytes` is **not** less than 65532):
        #    - MUST ignore the `ping`.
        if msg.fields["num_pong_bytes"] >= 65532:
            return []

        # A node sending a `pong` message:
        #  - SHOULD set `ignored` to 0s.
        #  - MUST NOT set `ignored` to sensitive data such as secrets or
        #    portions of initialized
        outmsg = Message(
            namespace().get_msgtype("pong"), ignored="00" * msg.fields["num_pong_bytes"]
        )
        return [outmsg]

    @staticmethod
    def ignore_gossip_queries(msg: Message) -> Optional[List[Message]]:
        """Ignore gossip_timestamp_filter, query_channel_range and query_short_channel_ids.  Respond to pings."""
        if msg.messagetype.name in (
            "gossip_timestamp_filter",
            "query_channel_range",
            "query_short_channel_ids",
        ):
            return []
        return ExpectMsg.ignore_pings(msg)

    @staticmethod
    def ignore_all_gossip(msg: Message) -> Optional[List[Message]]:
        """Ignore any gossip messages.  Respond to pings."""
        # BOLT #1: The messages are grouped logically into five
        # groups, ordered by the most significant bit that is set: ...
        #   - Routing (types `256`-`511`): messages containing node and channel
        #     announcements, as well as any active route exploration (described
        #     in [BOLT #7](07-routing-gossip.md))
        if msg.messagetype.number in range(256, 512):
            return []
        return ExpectMsg.ignore_pings(msg)

    def __init__(
        self,
        msgtypename: str,
        if_match: Callable[["ExpectMsg", Message, "Runner"], None] = _default_if_match,
        ignore: Optional[Callable[[Message], Optional[List[Message]]]] = None,
        connprivkey: Optional[str] = None,
        **kwargs: Union[str, Resolvable],
    ):
        super().__init__(connprivkey)
        self.msgtype = namespace().get_msgtype(msgtypename)
        if not self.msgtype:
            raise SpecFileError(self, "Unknown msgtype {}".format(msgtypename))
        self.kwargs = kwargs
        self.if_match = if_match
        # Assigning this in the __init__ line doesn't work!
        if ignore is None:
            ignore = self.ignore_gossip_queries
        self.ignore = ignore

    def message_match(self, runner: "Runner", msg: Message) -> Optional[str]:
        """Does this message match what we expect?"""
        partmessage = Message(self.msgtype, **self.resolve_args(runner, self.kwargs))

        ret = cmp_msg(msg, partmessage)
        if ret is None:
            self.if_match(self, msg, runner)
            msg_to_stash(runner, self, msg)
        return ret

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        conn = self.find_conn(runner)
        while True:
            binmsg = runner.get_output_message(conn, self)
            if binmsg is None:
                raise EventError(
                    self, f"Did not receive a message {self.msgtype} from runner"
                )

            for e in conn.must_not_events:
                if e.matches(binmsg):
                    raise EventError(
                        self, "Got msg banned by {}: {}".format(e, binmsg.hex())
                    )
            logging.debug(f"raw msg {binmsg}")
            # Might be completely unknown to namespace.
            try:
                msg = Message.read(namespace(), io.BytesIO(binmsg))
                runner.add_stash(msg.messagetype.name, msg)
            except ValueError as ve:
                raise EventError(
                    self, "Runner gave bad msg {}: {}".format(binmsg.hex(), ve)
                )
            logging.debug(f"decoded msg {msg.to_str()}")
            # Ignore function may tell us to respond.
            response = self.ignore(msg)
            if response is not None:
                for msg in response:
                    binm = io.BytesIO()
                    msg.write(binm)
                    runner.recv(self, conn, binm.getvalue())
                continue

            err = self.message_match(runner, msg)
            if err:
                raise EventError(self, "{}: message was {}".format(err, msg.to_str()))

            break
        return True


class Block(Event):
    """Generate a block, at blockheight, with optional txs."""

    def __init__(
        self,
        blockheight: int,
        number: Union[int, Callable] = 1,
        txs: List[ResolvableStr] = [],
    ):
        super().__init__()
        self.blockheight = blockheight
        self.number = number
        self.txs = txs

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        # Oops, did they ask us to produce a block with no predecessor?
        if runner.getblockheight() + 1 < self.blockheight:
            raise SpecFileError(
                self,
                "Cannot generate block #{} at height {}".format(
                    self.blockheight, runner.getblockheight()
                ),
            )

        # Throw away blocks we're replacing.
        if runner.getblockheight() >= self.blockheight:
            runner.trim_blocks(self.blockheight - 1)

        if isinstance(self.number, Callable):
            self.number = self.resolve_arg(None, runner, self.number)

        # Add new one
        runner.add_blocks(
            self, [self.resolve_arg("tx", runner, tx) for tx in self.txs], self.number
        )
        assert runner.getblockheight() == self.blockheight - 1 + self.number
        return True


class ExpectTx(Event):
    """Expect the runner to broadcast a transaction"""

    def __init__(self, txid: ResolvableStr):
        super().__init__()
        self.txid = txid

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        runner.expect_tx(self, self.resolve_arg("txid", runner, self.txid))
        return True


class FundChannel(PerConnEvent):
    """Tell the runner to fund a channel with this peer."""

    def __init__(
        self,
        amount: ResolvableInt,
        feerate: ResolvableInt = 253,
        expect_fail: ResolvableBool = False,
        connprivkey: Optional[str] = None,
    ):
        super().__init__(connprivkey)
        self.amount = amount
        self.feerate = feerate
        self.expect_fail = expect_fail

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        runner.fundchannel(
            self,
            self.find_conn(runner),
            self.resolve_arg("amount", runner, self.amount),
            self.resolve_arg("feerate", runner, self.feerate),
            self.resolve_arg("expect_fail", runner, self.expect_fail),
        )
        return True


class InitRbf(PerConnEvent):
    def __init__(
        self,
        channel_id: ResolvableStr,
        amount: ResolvableInt,
        utxo_tx: ResolvableStr,
        utxo_outnum: ResolvableInt,
        feerate: int,
        connprivkey: Optional[str] = None,
    ):
        super().__init__(connprivkey)
        self.channel_id = channel_id
        self.amount = amount
        self.feerate = feerate
        self.utxo_tx = utxo_tx
        self.utxo_outnum = utxo_outnum

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        utxo_tx = self.resolve_arg("utxo_tx", runner, self.utxo_tx)
        txid = CTransaction.deserialize(bytes.fromhex(utxo_tx)).GetTxid()[::-1].hex()

        runner.init_rbf(
            self,
            self.find_conn(runner),
            self.resolve_arg("channel_id", runner, self.channel_id),
            self.resolve_arg("amount", runner, self.amount),
            txid,
            self.resolve_arg("utxo_outnum", runner, self.utxo_outnum),
            self.feerate,
        )

        return True


class Invoice(Event):
    def __init__(self, amount: int, preimage: ResolvableStr):
        super().__init__()
        self.preimage = preimage
        self.amount = amount

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        runner.invoice(
            self,
            self.amount,
            check_hex(self.resolve_arg("preimage", runner, self.preimage), 64),
        )
        return True


class AddHtlc(PerConnEvent):
    def __init__(
        self, amount: int, preimage: ResolvableStr, connprivkey: Optional[str] = None
    ):
        super().__init__(connprivkey)
        self.preimage = preimage
        self.amount = amount

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        runner.addhtlc(
            self,
            self.find_conn(runner),
            self.amount,
            check_hex(self.resolve_arg("preimage", runner, self.preimage), 64),
        )
        return True


class DualFundAccept(Event):
    def __init__(self) -> None:
        super().__init__()

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        runner.accept_add_fund(self)
        return True


class ExpectError(PerConnEvent):
    def __init__(self, connprivkey: Optional[str] = None):
        super().__init__(connprivkey)

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        error = runner.check_error(self, self.find_conn(runner))
        if error is None:
            raise EventError(self, "No error found")
        if runner._is_dummy():
            return True
        error = bytes.fromhex(error)
        msg = Message.read(namespace(), io.BytesIO(error))
        logging.info(f"message received {msg.messagetype.name}, hex {error.hex()}")
        if msg.messagetype.name not in "error":
            raise EventError(
                self,
                f"not error found but received `{msg.messagetype.name}` with hex: `{error.hex()}`",
            )
        return True


class ExpectDisconnect(PerConnEvent):
    """This is considerer an hack, because the protocol
    is not specifing what fail the connection means, so in
    some case a node close the connection without any message
    or in a local env the connection got close before sending
    out the msg, so this event is the simplest way to work around
    this current issue."""

    def __init__(self, connprivkey: Optional[str] = None):
        super().__init__(connprivkey)

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        msg = runner.check_error(self, self.find_conn(runner))
        # in this case of the dummy runner we return a `Dummy error`
        # but in this case we wan receive an None value
        # because the connected got close before
        return msg is None or runner._is_dummy()


class CheckEq(Event):
    """Event to check a condition is true"""

    def __init__(self, a: Resolvable, b: Resolvable):
        super().__init__()
        self.a = a
        self.b = b

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        a = self.resolve_arg("a", runner, self.a)
        b = self.resolve_arg("b", runner, self.b)
        # dummy runner generates dummy fields.
        if a != b and not runner._is_dummy():
            raise EventError(self, "{} != {}".format(a, b))
        return True


class CloseChannel(Event):
    """Implementing the lnprototest event related to the
    close channel operation."""

    def __init__(self, channel_id: ResolvableStr):
        super(CloseChannel, self).__init__()
        self.channel_id = channel_id

    def action(self, runner: "Runner") -> bool:
        super().action(runner)
        channel_id = self.resolve_arg("channel_id", runner, self.channel_id)
        runner.close_channel(channel_id)
        return True


def msg_to_stash(runner: "Runner", event: Event, msg: Message) -> None:
    """ExpectMsg and Msg save every field to the stash, in order"""
    fields = msg.to_py()

    stash = runner.get_stash(event, type(event).__name__, [])
    stash.append((msg.messagetype.name, fields))
    runner.add_stash(type(event).__name__, stash)


def cmp_obj(obj: Any, expected: Any, prefix: str) -> Optional[str]:
    """Return None if every field in expected matches a field in obj.  Otherwise return a complaint"""
    if isinstance(expected, collections.abc.Mapping):
        for k, v in expected.items():
            if k not in obj:
                return "Missing field {}".format(prefix + "." + k)
            diff = cmp_obj(obj[k], v, prefix + "." + k)
            if diff:
                return diff
    elif not isinstance(expected, str) and isinstance(
        expected, collections.abc.Sequence
    ):
        # Should we allow expected to be shorter?
        if len(expected) != len(obj):
            return "Expected {} elements, got {} in {}: expected {} not {}".format(
                len(expected), len(obj), prefix, expected, obj
            )
        for i in range(len(expected)):
            diff = cmp_obj(obj[i], expected[i], "{}[{}]".format(prefix, i))
            if diff:
                return diff
    elif isinstance(expected, str) and expected.startswith("Sig("):
        # Special handling for signature comparisons.
        if Sig.from_str(expected) != Sig.from_str(obj):
            return "{}: signature mismatch {} != {}".format(prefix, obj, expected)
    else:
        if obj != expected:
            return "{}: {} != {}".format(prefix, obj, expected)

    return None


def cmp_msg(msg: Message, expected: Message) -> Optional[str]:
    """Return None if every field in expected matches a field in msg.  Otherwise return a complaint"""
    if msg.messagetype != expected.messagetype:
        return "Expected {}, got {}".format(expected.messagetype, msg.messagetype)

    obj = msg.to_py()
    expected_obj = expected.to_py()

    return cmp_obj(obj, expected_obj, expected.messagetype.name)


@overload
def msat(sats: int) -> int:
    ...


@overload
def msat(
    sats: Callable[["Runner", "Event", str], int]
) -> Callable[["Runner", "Event", str], int]:
    ...


def msat(sats: ResolvableInt) -> ResolvableInt:
    """Convert a field from statoshis to millisatoshis"""

    def _msat(runner: "Runner", event: Event, field: str) -> int:
        if callable(sats):
            return 1000 * sats(runner, event, field)
        else:
            return 1000 * sats

    if callable(sats):
        return _msat
    else:
        return 1000 * sats


def negotiated(
    a_features: ResolvableStr,
    b_features: ResolvableStr,
    included: List[int] = [],
    excluded: List[int] = [],
) -> ResolvableBool:
    def has_feature(fbit: int, featurebits: str) -> bool:
        # Feature bits go in optional/compulsory pairs.
        altfbit = fbit ^ 1
        return has_bit(featurebits, fbit) or has_bit(featurebits, altfbit)

    def _negotiated(runner: "Runner", event: Event, field: str) -> bool:
        a = event.resolve_arg("features", runner, a_features)
        b = event.resolve_arg("features", runner, b_features)

        for i in included:
            if not has_feature(i, a) or not has_feature(i, b):
                return False

        for e in excluded:
            if has_feature(e, a) or has_feature(e, b):
                return False

        return True

    return _negotiated



#! /usr/bin/python3
# #### Dummy runner which you should replace with real one. ####
import io
from .runner import Runner, Conn
from .event import Event, ExpectMsg, MustNotMsg
from typing import List, Optional
from .keyset import KeySet
from pyln.proto.message import (
    Message,
    FieldType,
    DynamicArrayType,
    EllipsisArrayType,
    SizedArrayType,
)
from typing import Any


class DummyRunner(Runner):
    def __init__(self, config: Any):
        super().__init__(config)

    def _is_dummy(self) -> bool:
        """The DummyRunner returns True here, as it can't do some things"""
        return True

    def get_keyset(self) -> KeySet:
        return KeySet(
            revocation_base_secret="11",
            payment_base_secret="12",
            htlc_base_secret="14",
            delayed_payment_base_secret="13",
            shachain_seed="FF" * 32,
        )

    def add_startup_flag(self, flag: str) -> None:
        if self.config.getoption("verbose"):
            print("[ADD STARTUP FLAG {}]".format(flag))
        return

    def get_node_privkey(self) -> str:
        return "01"

    def get_node_bitcoinkey(self) -> str:
        return "10"

    def has_option(self, optname: str) -> Optional[str]:
        return None

    def start(self) -> None:
        self.blockheight = 102

    def stop(self, print_logs: bool = False) -> None:
        pass

    def restart(self) -> None:
        super().restart()
        if self.config.getoption("verbose"):
            print("[RESTART]")
        self.blockheight = 102

    def connect(self, event: Event, connprivkey: str) -> None:
        if self.config.getoption("verbose"):
            print("[CONNECT {} {}]".format(event, connprivkey))
        self.add_conn(Conn(connprivkey))

    def getblockheight(self) -> int:
        return self.blockheight

    def trim_blocks(self, newheight: int) -> None:
        if self.config.getoption("verbose"):
            print("[TRIMBLOCK TO HEIGHT {}]".format(newheight))
        self.blockheight = newheight

    def add_blocks(self, event: Event, txs: List[str], n: int) -> None:
        if self.config.getoption("verbose"):
            print("[ADDBLOCKS {} WITH {} TXS]".format(n, len(txs)))
        self.blockheight += n

    def disconnect(self, event: Event, conn: Conn) -> None:
        super().disconnect(event, conn)
        if self.config.getoption("verbose"):
            print("[DISCONNECT {}]".format(conn))

    def recv(self, event: Event, conn: Conn, outbuf: bytes) -> None:
        if self.config.getoption("verbose"):
            print("[RECV {} {}]".format(event, outbuf.hex()))

    def fundchannel(
        self,
        event: Event,
        conn: Conn,
        amount: int,
        feerate: int = 253,
        expect_fail: bool = False,
    ) -> None:
        if self.config.getoption("verbose"):
            print(
                "[FUNDCHANNEL TO {} for {} at feerate {}. Expect fail? {}]".format(
                    conn, amount, feerate, expect_fail
                )
            )

    def init_rbf(
        self,
        event: Event,
        conn: Conn,
        channel_id: str,
        amount: int,
        utxo_txid: str,
        utxo_outnum: int,
        feerate: int,
    ) -> None:
        if self.config.getoption("verbose"):
            print(
                "[INIT_RBF TO {} (channel {}) for {} at feerate {}. {}:{}".format(
                    conn, channel_id, amount, feerate, utxo_txid, utxo_outnum
                )
            )

    def invoice(self, event: Event, amount: int, preimage: str) -> None:
        if self.config.getoption("verbose"):
            print("[INVOICE for {} with PREIMAGE {}]".format(amount, preimage))

    def accept_add_fund(self, event: Event) -> None:
        if self.config.getoption("verbose"):
            print("[ACCEPT_ADD_FUND]")

    def addhtlc(self, event: Event, conn: Conn, amount: int, preimage: str) -> None:
        if self.config.getoption("verbose"):
            print(
                "[ADDHTLC TO {} for {} with PREIMAGE {}]".format(conn, amount, preimage)
            )

    @staticmethod
    def fake_field(ftype: FieldType) -> str:
        if isinstance(ftype, DynamicArrayType) or isinstance(ftype, EllipsisArrayType):
            # Byte arrays are literal hex strings
            if ftype.elemtype.name == "byte":
                return ""
            return "[]"
        elif isinstance(ftype, SizedArrayType):
            # Byte arrays are literal hex strings
            if ftype.elemtype.name == "byte":
                return "00" * ftype.arraysize
            return (
                "["
                + ",".join([DummyRunner.fake_field(ftype.elemtype)] * ftype.arraysize)
                + "]"
            )
        elif ftype.name in (
            "byte",
            "u8",
            "u16",
            "u32",
            "u64",
            "tu16",
            "tu32",
            "tu64",
            "bigsize",
            "varint",
        ):
            return "0"
        elif ftype.name in ("chain_hash", "channel_id", "sha256"):
            return "00" * 32
        elif ftype.name == "point":
            return "038f1573b4238a986470d250ce87c7a91257b6ba3baf2a0b14380c4e1e532c209d"
        elif ftype.name == "short_channel_id":
            return "0x0x0"
        elif ftype.name == "signature":
            return "01" * 64
        else:
            raise NotImplementedError(
                "don't know how to fake {} type!".format(ftype.name)
            )

    def get_output_message(self, conn: Conn, event: ExpectMsg) -> Optional[bytes]:
        if self.config.getoption("verbose"):
            print("[GET_OUTPUT_MESSAGE {}]".format(conn))

        # We make the message they were expecting.
        msg = Message(event.msgtype, **event.resolve_args(self, event.kwargs))

        # Fake up the other fields.
        for m in msg.missing_fields():
            ftype = msg.messagetype.find_field(m.name)
            msg.set_field(m.name, self.fake_field(ftype.fieldtype))

        binmsg = io.BytesIO()
        msg.write(binmsg)
        return binmsg.getvalue()

    def expect_tx(self, event: Event, txid: str) -> None:
        if self.config.getoption("verbose"):
            print("[EXPECT-TX {}]".format(txid))

    def check_error(self, event: Event, conn: Conn) -> Optional[str]:
        super().check_error(event, conn)
        if self.config.getoption("verbose"):
            print("[CHECK-ERROR {}]".format(event))
        return "Dummy error"

    def check_final_error(
        self,
        event: Event,
        conn: Conn,
        expected: bool,
        must_not_events: List[MustNotMsg],
    ) -> None:
        pass

    def close_channel(self, channel_id: str) -> bool:
        if self.config.getoption("verbose"):
            print("[CLOSE-CHANNEL {}]".format(channel_id))
        return True

    def is_running(self) -> bool:
        return True

    def teardown(self):
        pass



#! /usr/bin/python3
import logging
import shutil
import tempfile

import coincurve
import functools

from .bitfield import bitfield
from .errors import SpecFileError
from .structure import Sequence
from .event import Event, MustNotMsg, ExpectMsg
from .utils import privkey_expand
from .keyset import KeySet
from abc import ABC, abstractmethod
from typing import Dict, Optional, List, Union, Any, Callable


class Conn(object):
    """Class for connections.  Details filled in by the particular runner."""

    def __init__(self, connprivkey: str):
        """Create a connection from a node with the given hex privkey: we use
        trivial values for private keys, so we simply left-pad with zeroes"""
        self.name = connprivkey
        self.connprivkey = privkey_expand(connprivkey)
        self.pubkey = coincurve.PublicKey.from_secret(self.connprivkey.secret)
        self.expected_error = False
        self.must_not_events: List[MustNotMsg] = []

    def __str__(self) -> str:
        return self.name


class Runner(ABC):
    """Abstract base class for runners.

    Most of the runner parameters can be extracted at runtime, but we do
    require that minimum_depth be 3, just for test simplicity.
    """

    def __init__(self, config: Any):
        self.config = config
        self.directory = tempfile.mkdtemp(prefix="lnpt-cl-")
        # key == connprivkey, value == Conn
        self.conns: Dict[str, Conn] = {}
        self.last_conn: Optional[Conn] = None
        self.stash: Dict[str, Dict[str, Any]] = {}
        self.logger = logging.getLogger(__name__)
        if self.config.getoption("verbose"):
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

    def _is_dummy(self) -> bool:
        """The DummyRunner returns True here, as it can't do some things"""
        return False

    def find_conn(self, connprivkey: Optional[str]) -> Optional[Conn]:
        # Default is whatever we specified last.
        if connprivkey is None:
            return self.last_conn
        if connprivkey in self.conns:
            self.last_conn = self.conns[connprivkey]
            return self.last_conn
        return None

    def add_conn(self, conn: Conn) -> None:
        self.conns[conn.name] = conn
        self.last_conn = conn

    def disconnect(self, event: Event, conn: Conn) -> None:
        if conn is None:
            raise SpecFileError(event, "Unknown conn")
        del self.conns[conn.name]
        self.check_final_error(event, conn, conn.expected_error, conn.must_not_events)

    def check_error(self, event: Event, conn: Conn) -> Optional[str]:
        conn.expected_error = True
        return None

    def post_check(self, sequence: Sequence) -> None:
        # Make sure no connection had an error.
        for conn_name in list(self.conns.keys()):
            logging.debug(
                f"disconnection connection with key={conn_name} and value={self.conns[conn_name]}"
            )
            self.disconnect(sequence, self.conns[conn_name])

    def restart(self) -> None:
        self.conns = {}
        self.last_conn = None
        self.stash = {}

    # FIXME: Why can't we use SequenceUnion here?
    def run(self, events: Union[Sequence, List[Event], Event]) -> None:
        sequence = Sequence(events)
        self.start()
        while True:
            all_done = sequence.action(self)
            self.post_check(sequence)
            if all_done:
                self.stop()
                return
            self.restart()

    def add_stash(self, stashname: str, vals: Any) -> None:
        """Add a dict to the stash."""
        self.stash[stashname] = vals

    def get_stash(self, event: Event, stashname: str, default: Any = None) -> Any:
        """Get an entry from the stash."""
        if stashname not in self.stash:
            if default is not None:
                return default
            raise SpecFileError(event, "Unknown stash name {}".format(stashname))
        return self.stash[stashname]

    def teardown(self):
        """The Teardown method is called at the end of the test,
        and it is used to clean up the root dir where the tests are run."""
        shutil.rmtree(self.directory)

    def runner_features(
        self,
        additional_features: Optional[List[int]] = None,
        globals: bool = False,
    ) -> str:
        """
        Provide the features required by the node.
        """
        if additional_features is None:
            return ""
        else:
            return bitfield(*additional_features)

    @abstractmethod
    def is_running(self) -> bool:
        """Return a boolean value that tells whether the runner is running
        or not.
        Is leave up to the runner implementation to keep the runner state"""
        pass

    @abstractmethod
    def connect(self, event: Event, connprivkey: str) -> None:
        pass

    @abstractmethod
    def check_final_error(
        self,
        event: Event,
        conn: Conn,
        expected: bool,
        must_not_events: List[MustNotMsg],
    ) -> None:
        pass

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self, print_logs: bool = False) -> None:
        """
        Stop the runner, and print all the log that the ln
        implementation produced.

        Print the log is useful when we have a failure e we need
        to debug what happens during the tests.
        """
        pass

    @abstractmethod
    def recv(self, event: Event, conn: Conn, outbuf: bytes) -> None:
        pass

    @abstractmethod
    def get_output_message(self, conn: Conn, event: ExpectMsg) -> Optional[bytes]:
        pass

    @abstractmethod
    def getblockheight(self) -> int:
        pass

    @abstractmethod
    def trim_blocks(self, newheight: int) -> None:
        pass

    @abstractmethod
    def add_blocks(self, event: Event, txs: List[str], n: int) -> None:
        pass

    @abstractmethod
    def expect_tx(self, event: Event, txid: str) -> None:
        pass

    @abstractmethod
    def invoice(self, event: Event, amount: int, preimage: str) -> None:
        pass

    @abstractmethod
    def accept_add_fund(self, event: Event) -> None:
        pass

    @abstractmethod
    def fundchannel(
        self,
        event: Event,
        conn: Conn,
        amount: int,
        feerate: int = 0,
        expect_fail: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def init_rbf(
        self,
        event: Event,
        conn: Conn,
        channel_id: str,
        amount: int,
        utxo_txid: str,
        utxo_outnum: int,
        feerate: int,
    ) -> None:
        pass

    @abstractmethod
    def addhtlc(self, event: Event, conn: Conn, amount: int, preimage: str) -> None:
        pass

    @abstractmethod
    def get_keyset(self) -> KeySet:
        pass

    @abstractmethod
    def get_node_privkey(self) -> str:
        pass

    @abstractmethod
    def get_node_bitcoinkey(self) -> str:
        pass

    @abstractmethod
    def has_option(self, optname: str) -> Optional[str]:
        pass

    @abstractmethod
    def add_startup_flag(self, flag: str) -> None:
        pass

    @abstractmethod
    def close_channel(self, channel_id: str) -> None:
        """
        Close the channel with the specified channel id.

        :param channel_id:  the channel id as string value where the
        caller want to close;
        :return No value in case of success is expected,
        but an `RpcError` is expected in case of err.
        """
        pass


def remote_revocation_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote revocation basepoint"""

    def _remote_revocation_basepoint(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().revocation_basepoint()

    return _remote_revocation_basepoint


def remote_payment_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote payment basepoint"""

    def _remote_payment_basepoint(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().payment_basepoint()

    return _remote_payment_basepoint


def remote_delayed_payment_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote delayed_payment basepoint"""

    def _remote_delayed_payment_basepoint(
        runner: Runner, event: Event, field: str
    ) -> str:
        return runner.get_keyset().delayed_payment_basepoint()

    return _remote_delayed_payment_basepoint


def remote_htlc_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote htlc basepoint"""

    def _remote_htlc_basepoint(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().htlc_basepoint()

    return _remote_htlc_basepoint


def remote_funding_pubkey() -> Callable[[Runner, Event, str], str]:
    """Get the remote funding pubkey (FIXME: we assume there's only one!)"""

    def _remote_funding_pubkey(runner: Runner, event: Event, field: str) -> str:
        return (
            coincurve.PublicKey.from_secret(
                privkey_expand(runner.get_node_bitcoinkey()).secret
            )
            .format()
            .hex()
        )

    return _remote_funding_pubkey


def remote_funding_privkey() -> Callable[[Runner, Event, str], str]:
    """Get the remote funding privkey (FIXME: we assume there's only one!)"""

    def _remote_funding_privkey(runner: Runner, event: Event, field: str) -> str:
        return runner.get_node_bitcoinkey()

    return _remote_funding_privkey


def remote_per_commitment_point(n: int) -> Callable[[Runner, Event, str], str]:
    """Get the n'th remote per-commitment point"""

    def _remote_per_commitment_point(
        n: int, runner: Runner, event: Event, field: str
    ) -> str:
        return runner.get_keyset().per_commit_point(n)

    return functools.partial(_remote_per_commitment_point, n)


def remote_per_commitment_secret(n: int) -> Callable[[Runner, Event, str], str]:
    """Get the n'th remote per-commitment secret"""

    def _remote_per_commitment_secret(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().per_commit_secret(n)

    return _remote_per_commitment_secret



"""
Utils module that implement common function used across lnprototest library.
"""
import string
import coincurve
import time
import typing
import logging
import traceback

from typing import Union, Sequence, List
from enum import IntEnum

from lnprototest.keyset import KeySet


class Side(IntEnum):
    local = 0
    remote = 1

    def __not__(self) -> "Side":
        if self == Side.local:
            return Side.remote
        return Side.local


def check_hex(val: str, digits: int) -> str:
    if not all(c in string.hexdigits for c in val):
        raise ValueError("{} is not valid hex".format(val))
    if len(val) != digits:
        raise ValueError("{} not {} characters long".format(val, digits))
    return val


def privkey_expand(secret: str) -> coincurve.PrivateKey:
    # Privkey can be truncated, since we use tiny values a lot.
    return coincurve.PrivateKey(bytes.fromhex(secret).rjust(32, bytes(1)))


def pubkey_of(privkey: str) -> str:
    """Return the public key corresponding to this privkey"""
    return (
        coincurve.PublicKey.from_secret(privkey_expand(privkey).secret).format().hex()
    )


def privkey_for_index(index: int = 0) -> str:
    from lnprototest.utils.bitcoin_utils import utxo

    _, _, _, privkey, _ = utxo(index)
    return privkey


def gen_random_keyset(counter: int = 20) -> KeySet:
    """Helper function to generate a random keyset."""

    from lnprototest import privkey_expand

    return KeySet(
        revocation_base_secret=f"{counter + 1}",
        payment_base_secret=f"{counter + 2}",
        htlc_base_secret=f"{counter + 3}",
        delayed_payment_base_secret=f"{counter + 4}",
        shachain_seed="00" * 32,
    )


def wait_for(success: typing.Callable, timeout: int = 180) -> None:
    start_time = time.time()
    interval = 0.25
    while not success():
        time_left = start_time + timeout - time.time()
        if time_left <= 0:
            raise ValueError("Timeout while waiting for {}", success)
        time.sleep(min(interval, time_left))
        interval *= 2
        if interval > 5:
            interval = 5


def get_traceback(e: Exception) -> str:
    lines = traceback.format_exception(type(e), e, e.__traceback__)
    return "".join(lines)


def run_runner(runner: "Runner", test: Union[Sequence, List["Event"], "Event"]) -> None:
    """
    The pytest using the assertion as safe failure, and the exception it is only
    an event that must not happen.

    From design, lnprototest fails with an exception, and for this reason, if the
    lnprototest throws an exception, we catch it, and we fail with an assent.
    """
    try:
        runner.run(test)
    except Exception as ex:
        runner.stop(print_logs=True)
        logging.error(get_traceback(ex))
        assert False, ex


def merge_events_sequences(
    pre: Union[Sequence, List["Event"], "Event"],
    post: Union[Sequence, List["Event"], "Event"],
) -> Union[Sequence, List["Event"], "Event"]:
    """Merge the two list in the pre-post order"""
    pre.extend(post)
    return pre



from .ln_spec_utils import (
    LightningUtils,
    connect_to_node_helper,
    open_and_announce_channel_helper,
)
from .utils import (
    Side,
    privkey_expand,
    wait_for,
    check_hex,
    gen_random_keyset,
    run_runner,
    pubkey_of,
    check_hex,
    privkey_for_index,
    merge_events_sequences,
)
from .bitcoin_utils import (
    ScriptType,
    BitcoinUtils,
    utxo,
    utxo_amount,
    funding_amount_for_utxo,
    tx_spendable,
    tx_out_for_index,
)



"""lnprototest: a framework and test suite for checking lightning spec protocol compliance.

This package is unusual, in that its main purpose is to carry the unit
tests, which can be run against a Lightning node implementation, using
an adapter called a 'Runner'.  Two runners are included: the
DummyRunner which is the default, and mainly useful to sanity check
the tests themselves, and clightning.Runner.

The documentation for the classes themselves should cover much of the
reference material, and the tutorial should get you started.

"""
from .errors import EventError, SpecFileError

from .event import (
    Event,
    Connect,
    Disconnect,
    Msg,
    RawMsg,
    ExpectMsg,
    MustNotMsg,
    Block,
    ExpectTx,
    FundChannel,
    InitRbf,
    Invoice,
    AddHtlc,
    CheckEq,
    ExpectError,
    ResolvableInt,
    ResolvableStr,
    Resolvable,
    ResolvableBool,
    msat,
    negotiated,
    DualFundAccept,
    Wait,
    CloseChannel,
    ExpectDisconnect,
)

from .structure import Sequence, OneOf, AnyOrder, TryAll

from .runner import (
    Runner,
    Conn,
    remote_revocation_basepoint,
    remote_payment_basepoint,
    remote_delayed_payment_basepoint,
    remote_htlc_basepoint,
    remote_per_commitment_point,
    remote_per_commitment_secret,
    remote_funding_pubkey,
    remote_funding_privkey,
)
from .dummyrunner import DummyRunner
from .namespace import (
    peer_message_namespace,
    namespace,
    assign_namespace,
    make_namespace,
)
from .bitfield import bitfield, has_bit, bitfield_len
from .signature import SigType, Sig
from .keyset import KeySet
from .commit_tx import Commit, HTLC, UpdateCommit
from .utils import (
    Side,
    privkey_expand,
    wait_for,
    LightningUtils,
    ScriptType,
    BitcoinUtils,
)
from .funding import (
    AcceptFunding,
    CreateFunding,
    CreateDualFunding,
    Funding,
    AddInput,
    AddOutput,
    FinalizeFunding,
    AddWitnesses,
)
from .proposals import dual_fund_csv, channel_type_csv

__all__ = [
    "EventError",
    "SpecFileError",
    "Resolvable",
    "ResolvableInt",
    "ResolvableStr",
    "ResolvableBool",
    "Event",
    "Connect",
    "Disconnect",
    "DualFundAccept",
    "CreateDualFunding",
    "AddInput",
    "AddOutput",
    "FinalizeFunding",
    "AddWitnesses",
    "Msg",
    "RawMsg",
    "ExpectMsg",
    "Block",
    "ExpectTx",
    "FundChannel",
    "InitRbf",
    "Invoice",
    "AddHtlc",
    "ExpectError",
    "Sequence",
    "OneOf",
    "AnyOrder",
    "TryAll",
    "CheckEq",
    "MustNotMsg",
    "SigType",
    "Sig",
    "DummyRunner",
    "Runner",
    "Conn",
    "KeySet",
    "peer_message_namespace",
    "namespace",
    "assign_namespace",
    "make_namespace",
    "bitfield",
    "has_bit",
    "bitfield_len",
    "msat",
    "negotiated",
    "remote_revocation_basepoint",
    "remote_payment_basepoint",
    "remote_delayed_payment_basepoint",
    "remote_htlc_basepoint",
    "remote_per_commitment_point",
    "remote_per_commitment_secret",
    "remote_funding_pubkey",
    "remote_funding_privkey",
    "Commit",
    "HTLC",
    "UpdateCommit",
    "Side",
    "AcceptFunding",
    "CreateFunding",
    "Funding",
    "regtest_hash",
    "privkey_expand",
    "Wait",
    "dual_fund_csv",
    "channel_type_csv",
    "wait_for",
    "CloseChannel",
    "ExpectDisconnect",
]



#!/usr/bin/python3
# This script exercises the core-lightning implementation

# Released by Rusty Russell under CC0:
# https://creativecommons.org/publicdomain/zero/1.0/

import os
import shutil
import subprocess
import logging
import socket

from contextlib import closing
from typing import Any, Callable
from bitcoin.rpc import RawProxy
from .backend import Backend


class BitcoinProxy:
    """Wrapper for BitcoinProxy to reconnect.

    Long wait times between calls to the Bitcoin RPC could result in
    `bitcoind` closing the connection, so here we just create
    throwaway connections. This is easier than to reach into the RPC
    library to close, reopen and reauth upon failure.
    """

    def __init__(self, btc_conf_file: str, *args: Any, **kwargs: Any):
        self.btc_conf_file = btc_conf_file

    def __getattr__(self, name: str) -> Callable:
        if name.startswith("__") and name.endswith("__"):
            # Python internal stuff
            raise AttributeError

        def f(*args: Any) -> Callable:
            self.__proxy = RawProxy(btc_conf_file=self.btc_conf_file)

            logging.debug(
                "Calling {name} with arguments {args}".format(name=name, args=args)
            )
            res = self.__proxy._call(name, *args)
            logging.debug("Result for {name} call: {res}".format(name=name, res=res))
            return res

        # Make debuggers show <function bitcoin.rpc.name> rather than <function
        # bitcoin.rpc.<lambda>>
        f.__name__ = name
        return f


class Bitcoind(Backend):
    """Starts regtest bitcoind on an ephemeral port, and returns the RPC proxy"""

    def __init__(self, basedir: str, with_wallet: bool = True):
        self.with_wallet = with_wallet
        self.rpc = None
        self.proc = None
        self.base_dir = basedir
        logging.debug(f"Base dir is {basedir}")
        self.bitcoin_dir = os.path.join(basedir, "bitcoind")
        self.bitcoin_conf = os.path.join(self.bitcoin_dir, "bitcoin.conf")
        self.cmd_line = [
            "bitcoind",
            "-datadir={}".format(self.bitcoin_dir),
            "-server",
            "-regtest",
            "-logtimestamps",
            "-nolisten",
        ]
        self.btc_version = None

    def __reserve(self) -> int:
        """
        When python asks for a free port from the os, it is possible that
        with concurrent access, the port that is picked is a port that is not free
        anymore when we go to bind the daemon like bitcoind port.

        Source: https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number
        """
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def __init_bitcoin_conf(self):
        """Init the bitcoin core directory with all the necessary information
        to startup the node"""
        if not os.path.exists(self.bitcoin_dir):
            os.makedirs(self.bitcoin_dir)
            logging.debug(f"Creating {self.bitcoin_dir} directory")
        self.port = self.__reserve()
        logging.debug("Port is {}, dir is {}".format(self.port, self.bitcoin_dir))
        # For after 0.16.1 (eg. 3f398d7a17f136cd4a67998406ca41a124ae2966), this
        # needs its own [regtest] section.
        logging.debug(f"Writing bitcoin conf file at {self.bitcoin_conf}")
        with open(self.bitcoin_conf, "w") as f:
            f.write("regtest=1\n")
            f.write("rpcuser=rpcuser\n")
            f.write("rpcpassword=rpcpass\n")
            f.write("[regtest]\n")
            f.write("rpcport={}\n".format(self.port))
        self.rpc = BitcoinProxy(btc_conf_file=self.bitcoin_conf)

    def __version_compatibility(self) -> None:
        """
        This method tries to manage the compatibility between
        different versions of Bitcoin Core implementation.

        This method could sometimes be useful when it is necessary to
        run the test with a different version of Bitcoin core.
        """
        if self.rpc is None:
            # Sanity check
            raise ValueError("bitcoind not initialized")

        self.btc_version = self.rpc.getnetworkinfo()["version"]
        assert self.btc_version is not None
        logging.debug("Bitcoin Core version {}".format(self.btc_version))
        if self.btc_version >= 210000:
            # Maintains the compatibility between wallet
            # different ln implementation can use the main wallet (?)
            if self.with_wallet:
                self.rpc.createwallet("main")  # Automatically loads

    def __is__bitcoind_ready(self) -> bool:
        """Check if bitcoind is ready during the execution"""
        if self.proc is None:
            # Sanity check
            raise ValueError("bitcoind not initialized")

        # Wait for it to startup.
        while b"Done loading" not in self.proc.stdout.readline():
            pass
        return True

    def start(self) -> None:
        if self.rpc is None:
            self.__init_bitcoin_conf()
        # TODO: We can move this to a single call and not use Popen
        self.proc = subprocess.Popen(self.cmd_line, stdout=subprocess.PIPE)
        assert self.proc.stdout

        # Wait for it to startup.
        while not self.__is__bitcoind_ready():
            logging.debug("Bitcoin core is loading")

        self.__version_compatibility()
        # Block #1.
        # Privkey the coinbase spends to:
        #    cUB4V7VCk6mX32981TWviQVLkj3pa2zBcXrjMZ9QwaZB5Kojhp59
        self.rpc.submitblock(
            "0000002006226e46111a0b59caaf126043eb5bbf28c34f3a5e332a1fc7b2b73cf188910f84591a56720aabc8023cecf71801c5e0f9d049d0c550ab42412ad12a67d89f3a3dbb6c60ffff7f200400000001020000000001010000000000000000000000000000000000000000000000000000000000000000ffffffff03510101ffffffff0200f2052a0100000016001419f5016f07fe815f611df3a2a0802dbd74e634c40000000000000000266a24aa21a9ede2f61c3f71d1defd3fa999dfa36953755c690689799962b48bebd836974e8cf90120000000000000000000000000000000000000000000000000000000000000000000000000"
        )
        self.rpc.generatetoaddress(100, self.rpc.getnewaddress())

    def stop(self) -> None:
        self.rpc.stop()
        self.proc.kill()
        shutil.rmtree(os.path.join(self.bitcoin_dir, "regtest"))

    def restart(self) -> None:
        # Only restart if we have to.
        if self.rpc.getblockcount() != 101 or self.rpc.getrawmempool() != []:
            self.stop()
            self.start()
